/*********************************************************************************
 * Copyright (c) 2006-2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************************/


package de.fzj.unicore.xuudb.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Properties;

import javax.security.auth.x500.X500Principal;

import org.apache.log4j.Logger;

import de.fzJuelich.unicore.xuudb.AddCertificateDocument;
import de.fzJuelich.unicore.xuudb.AddCertificateResponseDocument;
import de.fzJuelich.unicore.xuudb.CheckCertificateDocument;
import de.fzJuelich.unicore.xuudb.CheckCertificateResponseDocument;
import de.fzJuelich.unicore.xuudb.CheckDNDataType;
import de.fzJuelich.unicore.xuudb.CheckDNDocument;
import de.fzJuelich.unicore.xuudb.CheckDNResponseDocument;
import de.fzJuelich.unicore.xuudb.CheckDataType;
import de.fzJuelich.unicore.xuudb.DatabaseType;
import de.fzJuelich.unicore.xuudb.ImportDatabaseDocument;
import de.fzJuelich.unicore.xuudb.ImportDatabaseResponseDocument;
import de.fzJuelich.unicore.xuudb.ListDatabaseDocument;
import de.fzJuelich.unicore.xuudb.ListDatabaseResponseDocument;
import de.fzJuelich.unicore.xuudb.LoginDataType;
import de.fzJuelich.unicore.xuudb.RemoveCertificateDocument;
import de.fzJuelich.unicore.xuudb.RemoveCertificateResponseDocument;
import de.fzJuelich.unicore.xuudb.UpdateCertificateDocument;
import de.fzJuelich.unicore.xuudb.UpdateCertificateResponseDocument;
import de.fzJuelich.unicore.xuudb.UpdateDataType;
import de.fzj.unicore.xuudb.Configuration;
import de.fzj.unicore.xuudb.X509Utils;
import de.fzj.unicore.xuudb.XUUDBInit;
import de.fzj.unicore.xuudb.interfaces.IAdmin;
import de.fzj.unicore.xuudb.interfaces.IPublic;

/**
 * execute admin commands
 * 
 * @author tweddell
 */ 
public class AdminController extends XUUDBInit {

	private static final Logger logger = Configuration.getLogger();

	private IAdmin admin;
	private IPublic query;

	private boolean batch=false;

	protected void initEndpoints(){
		ServiceFactory serviceFac = ServiceFactory.getInstance(AdminController.getConfig());
		if(admin==null){
			admin=serviceFac.getAdminEndpoint();
		}
		if(query==null){
			query=serviceFac.getPublicEndpoint();
		}
	}

	public AdminController(Properties props)throws Exception{
		super(props);
		checkBatch();
	}

	public AdminController(String cf)throws Exception{
		super(cf);
		checkBatch();
	}

	private void checkBatch(){
		try{
			String batchProp=getConfig().getProperty("xuudb.batch");
			if(batchProp==null){
				batchProp=System.getProperty("xuudb.batch", "false");
			}
			batch = Boolean.parseBoolean(batchProp);
		}catch(Exception ex){}
	}

	public void run(String[] args) throws Exception {
		String command = args[0];
		String[] shiftedArgs = new String[args.length-1];
		System.arraycopy(args, 1, shiftedArgs, 0, args.length-1);
		if(command.equalsIgnoreCase("help") || command.equalsIgnoreCase("-h")  || command.equalsIgnoreCase("--help")){
			XUUDBClient.printUsage();
			return;
		}

		initEndpoints();

		if(command.equalsIgnoreCase("add")) {
			this.add(shiftedArgs);
		} else if(command.equalsIgnoreCase("adddn")) {
			this.adddn(shiftedArgs);
		} else if(command.equalsIgnoreCase("remove")) {
			this.remove(shiftedArgs);
		} else if(command.equalsIgnoreCase("update")) {
			this.update(shiftedArgs);
		} else if(command.equalsIgnoreCase("list")) {
			this.list(shiftedArgs);
		} else if(command.equalsIgnoreCase("export")) {
			this.export_csv(shiftedArgs);
		} else if(command.equalsIgnoreCase("import")) {
			this.import_csv(shiftedArgs);
		} else if(command.equalsIgnoreCase("check-cert")) {
			this.testcheck(shiftedArgs);
		} else if(command.equalsIgnoreCase("check-dn")) {
			this.testcheckdn(shiftedArgs);
		} else {
			logger.error("Unknown Command: " + command);
			XUUDBClient.printUsage();
		}
	}

	private void testcheck(String[] args) {
		logger.info("Command: check-cert");
		for (int i = 0; i < args.length; i++) {
			System.out.println("Parameter " + i + ": " + args[i]);
		}
		if(args.length != 2) {
			exitWithError("Wrong argument count. Must be two. gcid and pemfile", null);
		}
		// read pem-file to cert
		String cert=null;
		try {
			System.out.println("Reading cert from file: " + args[1]);
			cert = X509Utils.getStringFromPEMFile(new File(args[1]));
		} catch (FileNotFoundException e) {
			exitWithError("Can't read certificate <" + args[1]+">", e);                                            
		}
		// preparing webservice call		
		CheckCertificateDocument outXML = CheckCertificateDocument.Factory.newInstance();
		CheckDataType checkDataTypeXml = CheckDataType.Factory.newInstance();
		checkDataTypeXml.setGcID(args[0]);
		checkDataTypeXml.setCertInPEM(cert);
		outXML.setCheckCertificate(checkDataTypeXml);
		// calling ws
		System.out.println("Invoking ...");
		CheckCertificateResponseDocument inXML = null;
		try {
			inXML = query.checkCertificate(outXML);
		} catch (Exception xfe) {
			logger.error("Error while invoking service: " + xfe.getMessage(),xfe);
			System.out.println("Error while invoking service: " + xfe.getMessage());
			return;
		}
		System.out.println("Done. Received: gcid=" + inXML.getCheckCertificateResponse().getGcID());
		System.out.println("Done. Received: token=" + inXML.getCheckCertificateResponse().getToken());
		System.out.println("Done. Received: xlogin=" + inXML.getCheckCertificateResponse().getXlogin());
		System.out.println("Done. Received: role=" + inXML.getCheckCertificateResponse().getRole());
		System.out.println("Done. Received: projects=" + inXML.getCheckCertificateResponse().getProjects());
	}

	private void testcheckdn(String[] args) {
		logger.info("Command: check-dn");
		for (int i = 0; i < args.length; i++) {
			System.out.println("Parameter " + i + ": " + args[i]);
		}
		if(args.length != 2) {
			exitWithError("Wrong argument count. Must be two. gcid and dn", null);
		}

		String dn=null;
		try {
			dn = new X500Principal(args[1]).getName();
		} catch (IllegalArgumentException e) {
			exitWithError("DN format error", e);
		}
		// preparing webservice call		

		CheckDNDocument outXML = CheckDNDocument.Factory.newInstance();
		CheckDNDataType checkDNDataTypeXml = CheckDNDataType.Factory.newInstance();
		checkDNDataTypeXml.setGcID(args[0]);
		checkDNDataTypeXml.setDistinguishedName(dn);
		outXML.setCheckDN(checkDNDataTypeXml);
		// calling ws
		System.out.println("Invoking ...");
		CheckDNResponseDocument inXML = null;
		try {
			inXML = query.checkDN(outXML);
		} catch (Exception xfe) {
			exitWithError("Error invoking service", xfe);
		}
		System.out.println("Done. Received: gcid=" + inXML.getCheckDNResponse().getGcID());
		System.out.println("Done. Received: token=" + inXML.getCheckDNResponse().getToken());
		System.out.println("Done. Received: xlogin=" + inXML.getCheckDNResponse().getXlogin());
		System.out.println("Done. Received: role=" + inXML.getCheckDNResponse().getRole());
		System.out.println("Done. Received: projects=" + inXML.getCheckDNResponse().getProjects());
	}

	private void adddn(String[] args) {
		logger.info("Command: adddn ");

		Boolean isDN=isDNMode();
		if(isDN!=null){
			if(!isDN.booleanValue()){
				exitWithError("Cannot add. The server is not running in DN mode. Please use the 'add' command.", null);
			}
		}
		for (int i = 0; i < args.length; i++) {
			logger.info("Parameter " + i + ": " + args[i]);
		}
		if(args.length != 5 && args.length != 4) {
			XUUDBClient.printUsage();
			return;
		}

		X500Principal dn=null;
		try {
			dn=new X500Principal(args[1]);
		} catch (Exception e) {
			exitWithError("Cannot add. DN format error?", e);
		}

		// preparing webservice call
		logger.info("Creating AddCertificateDocument");
		AddCertificateDocument outXML = AddCertificateDocument.Factory.newInstance();
		LoginDataType addUloginXML = LoginDataType.Factory.newInstance();
		addUloginXML.setGcID(args[0]);
		addUloginXML.setToken(dn.getName(X500Principal.RFC2253));
		addUloginXML.setXlogin(args[2]);
		addUloginXML.setRole(args[3]);
		if(args.length==5)
			addUloginXML.setProjects(args[4]);
		else
			addUloginXML.setProjects("");
		outXML.setAddCertificate(addUloginXML);

		// calling ws
		logger.info("Invoking ...");
		AddCertificateResponseDocument inXML = null;
		try {
			inXML = admin.addCertificate(outXML);
		} catch (Exception xfe) {
			exitWithError("Error while invoking service",xfe);
		}
		if(inXML.getAddCertificateResponse().indexOf("constraint") != -1 ) {
			exitWithError("Error! This dataset already exists. Constraint: GCID and DN must be unique",null);
		}
		System.out.println("Done. Received: " + inXML.getAddCertificateResponse());
		logger.info("Done. Received: " + inXML.getAddCertificateResponse());

	}

	private void export_csv(String[] args) {
		logger.info("Command: export ");
		for (int i = 0; i < args.length; i++) {
			logger.info("Parameter " + i + ": " + args[i]);
		}
		if(args.length >2 || args.length == 0) {
			XUUDBClient.printUsage();
			return;
		}
		String fname = args[0];
		if (!fname.endsWith(".csv"))
			fname+=".csv";			
		boolean force=args.length>1 && "force".equalsIgnoreCase(args[1]);

		File exp = new File(fname);
		PrintStream writer=null;
		try {
			FileOutputStream fop;
			if(exp.createNewFile()==false && !force) {
				System.out.println("File exists: " + exp.getAbsolutePath());
				System.out.println("Use the 'force' parameter to override.");
				return;
			}
			fop = new FileOutputStream(exp);
			writer = new PrintStream(fop, false);

		} catch (IOException e) {
			exitWithError("Cannot write to file: " + exp.getAbsolutePath(), e);
		}

		ListDatabaseDocument outXML = ListDatabaseDocument.Factory.newInstance();
		LoginDataType data0 = LoginDataType.Factory.newInstance();
		outXML.setListDatabase(data0);

		logger.info("Invoking ...");
		System.out.println("Invoking ...");
		ListDatabaseResponseDocument inXML = null;
		try {
			inXML =	admin.listDatabase(outXML);
		} catch (Exception xfe) {
			exitWithError("Error invoking service", xfe);
		}
		LoginDataType[] data = inXML.getListDatabaseResponse().getDatabaseArray();
		if(data==null)
			System.out.println("The database query result is empty.");
		else {
			System.out.println("The database query result contains " +
					inXML.getListDatabaseResponse().getDatabaseArray().length +
			" entries:");
			writer.println("Nr.;GcID;Xlogin;Role;Projects;CertInPEM");
			for (int i = 0;  i < data.length; i++) {
				writer.println((i+1) + ";" + 
						data[i].getGcID() + ";" + 
						data[i].getXlogin() + ";" +
						data[i].getRole()+ ";" +
						data[i].getProjects() + ";" +
						data[i].getToken());

			}
			writer.close();
			System.out.println("Data written to: " + exp.getAbsolutePath());
		}
	}

	private void list(String[] args) {
		logger.info("Command: list ");
		if(args.length>0){
			for (int i = 0; i < args.length; i++) {
				logger.info("Parameter " + i + ": " + args[i]);
			}
		}
		ListDatabaseDocument outXML = ListDatabaseDocument.Factory.newInstance();
		//prepare statement 
		try {
			logger.info("Building Ulogin for db query");
			LoginDataType data = this.parseArgsToUlogin(args);
			outXML.setListDatabase(data);
		} catch (FileNotFoundException e) {
			exitWithError("Cannot build query (check pem file path?)", e);
		} catch (Exception e) {
			exitWithError("Cannot build query (check parameters?)", e);
		}

		logger.info("Invoking ...");
		ListDatabaseResponseDocument inXML = null;
		try {
			inXML = admin.listDatabase(outXML);
		} catch (Exception xfe) {
			exitWithError("Error invoking service", xfe);
		}
		String info=inXML.getListDatabaseResponse().getXUUDBInfo();
		if(info!=null){
			System.out.println("XUUDB info: "+info);
		}
		LoginDataType[] data = inXML.getListDatabaseResponse().getDatabaseArray();
		if(data==null)
			System.out.println("The database query result is empty.");
		else {
			System.out.println("The database query result contains   " +
					inXML.getListDatabaseResponse().getDatabaseArray().length +
			"   entries:");
			System.out.printf("%4s|%-20s|%-15s|%-10s|%-20s|%-20s\n", "Nr", "GcID", "Xlogin", "Role", "Projects", "DN");
			System.out.println("---------------------------------------------------------------------------");
			for (int i = 0;  i < data.length; i++) {
				String dn = "Error resolving DN :/";
				try {
					X500Principal p=new X500Principal(data[i].getToken());
					dn = p.toString(); //user-friendly form, not the one really stored in XUUDB
				} catch (Exception pe) {
					try {
						X509Certificate cert = X509Utils.getX509FromPEMString(data[i].getToken());
						dn = cert.getSubjectDN().toString();
					} catch (CertificateException e) {	}
				}
				System.out.printf( "%4d|%-20s|%-15s|%-10s|%-20s|%-20s\n", 
						i+1,
						data[i].getGcID(),
						data[i].getXlogin(), 
						data[i].getRole(),
						data[i].getProjects(),
						dn);
			}
		}
	}

	private LoginDataType parseArgsToUlogin(String[] args) throws Exception {
		LoginDataType ret = LoginDataType.Factory.newInstance();
		boolean alreadyHaveToken=false;
		for(int i=0; i<args.length; ++i) {
			if(args[i].toLowerCase().startsWith("gcid="))
				ret.setGcID(args[i].substring(5));
			else if(args[i].toLowerCase().startsWith("xlogin="))
				ret.setXlogin(args[i].substring(7));
			else if(args[i].toLowerCase().startsWith("pemfile=")) {
				File pem = new File(args[i].substring(8));
				ret.setToken(X509Utils.getStringFromPEMFile(pem));
				alreadyHaveToken=true;
			}
			else if(args[i].toLowerCase().startsWith("dn=")) {
				if( alreadyHaveToken ) {
					throw new Exception("You can not use pemfile= and dn= at a time.");
				}
				try{
					String dn=new X500Principal(args[i].substring(3)).getName();
					ret.setToken(dn);
				}catch(Exception ex){
					throw new IllegalArgumentException("Not a valid DN: "+args[i].substring(3));
				}
			}
			else if(args[i].toLowerCase().startsWith("role="))
				ret.setRole(args[i].substring(5));
			else if(args[i].toLowerCase().startsWith("project="))
				ret.setProjects(args[i].substring(8));
			else {
				throw new Exception("Unknown parameter: " + args[i]);
			}
		}
		return ret;
	}

	private void update(String[] args) {
		logger.info("Command: update ");
		for (int i = 0; i < args.length; i++) {
			logger.info("Parameter " + i + ": " + args[i]);
		}
		if(args.length < 3) {
			XUUDBClient.printUsage();
			return;
		}

		// get args:
		String[] lastArgs = new String[args.length-2];
		for(int i=2;i<args.length;++i) {
			lastArgs[i-2] = args[i];
		}
		String gcid = args[0];
		String token = null;
		try {
			logger.debug("Trying to read token as a DN...");
			token = new X500Principal(args[1]).getName();
		} catch (Exception e) {
			logger.debug("Token is not a DN.");
			try {
				logger.debug("Trying to read token from file (x509)...");
				token = X509Utils.getStringFromPEMFile(new File(args[1]));
			} catch (FileNotFoundException e1) {
				exitWithError("Can't read pem from file",e1);
			}
		}

		//build outgoing request
		UpdateCertificateDocument outXMLDoc = UpdateCertificateDocument.Factory.newInstance();
		UpdateDataType outXML=outXMLDoc.addNewUpdateCertificate();
		outXML.setGcID(gcid);
		outXML.setToken(token);
		try {
			logger.info("Building Ulogin for db update");
			LoginDataType data = this.parseArgsToUlogin(lastArgs);
			outXML.setData(data);
		} catch (FileNotFoundException e) {
			exitWithError("Can't build query (check pem file path?)",e);
		} catch (Exception e) {
			exitWithError("Can't build query (check parameters?)",e);
		}

		logger.info("Invoking ...");
		System.out.println("Invoking ...");
		UpdateCertificateResponseDocument inXML = null;
		try {
			inXML =	admin.updateCertificate(outXMLDoc);
		} catch (Exception xfe) {
			exitWithError("Error invoking service",xfe);
		}
		logger.info("Done. Received: " + inXML.getUpdateCertificateResponse());
		System.out.println("Done. Received: " + inXML.getUpdateCertificateResponse());
	}

	private void remove(String[] args) {
		logger.info("Command: remove ");
		for (int i = 0; i < args.length; i++) {
			logger.info("Parameter " + i + ": " + args[i]);
		}
		if(args.length==0) {
			XUUDBClient.printUsage();
			return;
		}

		RemoveCertificateDocument outXML = RemoveCertificateDocument.Factory.newInstance();
		//prepare statement 
		if(args[0].equalsIgnoreCase("ALL")) {
			logger.info("Clearing complete database.");
			String msg="All entries will be removed!";
			if (!batch) {
				if (!confirm(msg)) {
					logger.info("Operation stopped by user");
					System.out.println("Database was not changed.");
					return;
				}
			}
			LoginDataType data = LoginDataType.Factory.newInstance();
			outXML.setRemoveCertificate(data);
		} else {
			try {
				logger.info("Building Ulogin for db query");
				LoginDataType data = this.parseArgsToUlogin(args);
				outXML.setRemoveCertificate(data);
			} catch (FileNotFoundException e) {
				exitWithError("Can't build query (check pem file path?)",e);
			} catch (Exception e) {
				exitWithError("Can't build query (check parameters?)",e);
			}
		}

		logger.info("Invoking ...");
		System.out.println("Invoking ...");
		RemoveCertificateResponseDocument inXML = null;
		try {
			inXML =	admin.removeCertificate(outXML);
		} catch (Exception xfe) {
			exitWithError("Error invoking service", xfe);
		}
		logger.info("Done. Received: " + inXML.getRemoveCertificateResponse());
		System.out.println("Done. Received: " + inXML.getRemoveCertificateResponse());
	}

	private boolean confirm(String message){
		System.out.println(message+"\nAre you sure? [yes, no]");
		BufferedReader br = new BufferedReader(new InputStreamReader(System.in));
		String ans = "no";
		try {
			ans = br.readLine();
		} catch (Exception e) {}

		return "yes".equalsIgnoreCase(ans);
	}

	private void add(String[] args) {
		logger.info("Command: add ");
		for (int i = 0; i < args.length; i++) {
			logger.info("Parameter " + i + ": " + args[i]);
		}
		if(args.length != 5 && args.length != 4) {
			XUUDBClient.printUsage();
			return;
		}
		// read pem-file to cert
		String cert=null;
		try {
			logger.info("Reading cert from file: " + args[1]);
			cert = X509Utils.getStringFromPEMFile(new File(args[1]));
		} catch (FileNotFoundException e) {
			exitWithError("Can't read certificate file: " + args[1],e);
		}

		// check validity of cert
		logger.info("Checking validity from : " + args[1]);
		System.out.println("Checking validity from : " + args[1]);
		X509Certificate x509 = null;
		try {
			x509 = X509Utils.getX509FromPEMString(cert);
		} catch (CertificateException ce) {
			exitWithError("Can't build certificate (not a pem file?)", ce);
		}
		try {
			System.out.println("Issuer DN          \t" + x509.getIssuerDN());
			System.out.println("Not after          \t" + x509.getNotAfter());
			System.out.println("Not before         \t" + x509.getNotBefore());
			System.out.println("Serial No.         \t" + x509.getSerialNumber());
			System.out.println("Signature Alg.     \t" + x509.getSigAlgName());
			System.out.println("Signature Alg. OID \t" + x509.getSigAlgOID());
			System.out.println("Subject DN         \t" + x509.getSubjectDN());
			x509.checkValidity();
		} catch (CertificateExpiredException cee) {
			String msg="Certificate is expired: " + x509.getNotAfter();
			if(batch){
				System.out.println("Warning: " + msg );
			}
			else if(!confirm(msg)){
				exitWithError("Error: "+msg,cee);
			}
		} catch (CertificateNotYetValidException cnyve) {
			String msg="Certificate is invalid, it will be valid from: " + x509.getNotBefore();
			if(batch){
				System.out.println("Warning: " + msg );
			}
			else if(!confirm(msg)){
				exitWithError("Error: "+msg,cnyve);
			}
		}

		System.out.println("GCID     : " + args[0] );
		System.out.println("Role     : " + args[3] );
		System.out.println("Xlogin   : " + args[2] );
		if(args.length==5)
			System.out.println("Projects : " + args[4]);
		else 
			System.out.println("Projects :");

		if(!batch){
			// ask admin to add cert
			String msg="Do you really want to add this certificate?";
			if( !confirm(msg) ) {
				exitWithError("Aborted by user", null);
			}
		}
		// preparing webservice call
		logger.info("Creating AddCertificateDocument");
		AddCertificateDocument outXML = AddCertificateDocument.Factory.newInstance();
		LoginDataType addUloginXML = LoginDataType.Factory.newInstance();
		addUloginXML.setGcID(args[0]);
		addUloginXML.setToken(cert);
		addUloginXML.setRole(args[3]);
		addUloginXML.setXlogin(args[2]);
		if(args.length==5)
			addUloginXML.setProjects(args[4]);
		else
			addUloginXML.setProjects("");
		outXML.setAddCertificate(addUloginXML);

		// calling ws
		logger.info("Invoking ...");
		System.out.println("Invoking ...");
		AddCertificateResponseDocument inXML = null;
		try {
			inXML = admin.addCertificate(outXML);
		} catch (Exception xfe) {
			exitWithError("Error invoking service",xfe);
		}
		logger.info("Done. Received: " + inXML.getAddCertificateResponse());
		System.out.println("Done. Received: " + inXML.getAddCertificateResponse());
		if(inXML.getAddCertificateResponse().indexOf("Unique constraint violation") != -1 ) {
			exitWithError("This dataset already exists. Constraint: GCID and Certificate must be unique",null);
		}

	}


	private void import_csv(String[] args) {
		logger.info("Command: import ");
		for (int i = 0; i < args.length; i++) {
			logger.info("Parameter " + i + ": " + args[i]);
		}
		if(args.length <1 && args.length > 2) {
			XUUDBClient.printUsage();
			return;
		}
		//check if we should clear the DB before importing 
		boolean clear=false;
		if(args.length>1){
			clear="clearDB".equalsIgnoreCase(args[1]);
		}
		File incsv = new File(args[0]);
		if( !incsv.exists() ) {
			exitWithError("File does not exist: " + incsv.getAbsolutePath(),null);
		}
		BufferedReader br = null;
		try {
			br = new BufferedReader(new InputStreamReader(new FileInputStream(incsv)));
		} catch (FileNotFoundException e) {
			exitWithError("Error opening: " + incsv.getAbsolutePath(),e);
		}
		//using ListDatabaseResponseDocument as it can contain more LoginDataTypes
		ImportDatabaseDocument ldd = ImportDatabaseDocument.Factory.newInstance();
		DatabaseType dt = DatabaseType.Factory.newInstance();
		if(clear)dt.setClean(clear);
		ArrayList<LoginDataType> data = new ArrayList<LoginDataType>();
		try {
			br.readLine(); // read header line
			String line = null;
			int linec=0;
			int addc=0;
			while( (line = br.readLine()) != null ) {
				linec++; // first line is 1
				LoginDataType login = LoginDataType.Factory.newInstance();
				String[] tokens = line.split(";");
				logger.debug("Inspecting line: " + line );
				if( tokens.length == 6 ) {
					logger.debug(" >> Action: Adding to import queue at nr" + ++addc);
					// gcid
					login.setGcID(tokens[1]);
					// xlogin
					login.setXlogin(tokens[2]);
					//role
					login.setRole(tokens[3]);
					// projects
					login.setProjects(tokens[4]);
					// sectoken
					login.setToken(processToken(tokens[5]));
					data.add(login);
				} else {
					logger.warn("  >> Action: Ignoring line " +linec+".  Count of token is not 6");
				}
			}

		} catch (IOException e) {
			e.printStackTrace();
		}
		// converting arraylist to array
		LoginDataType[] ldt = new LoginDataType[data.size()];
		for(int i=0; i<ldt.length; i++) {
			ldt[i] = data.get(i);
		}
		dt.setDatabaseArray(ldt);
		ldd.setImportDatabase(dt);

		ImportDatabaseDocument outXML = ldd;

		logger.info("Invoking ...");
		ImportDatabaseResponseDocument inXML = null;
		try {
			inXML = admin.importDatabase(outXML);
		} catch (Exception xfe) {
			exitWithError("Error while invoking service",xfe);
		}		
		logger.info("Done. Received: " + inXML.getImportDatabaseResponse());
		System.out.println("Done. Received: " + inXML.getImportDatabaseResponse());

	}

	//in case token is a X500Name, canonicalize it
	private String processToken(String token){
		try{
			X500Principal dn=new X500Principal(token);
			return dn.getName(X500Principal.RFC2253);
		}catch(Exception ignored){}
		return token;
	}

	public void setAdminInterface(IAdmin admin) {
		this.admin = admin;
	}

	public void setQueryInterface(IPublic query) {
		this.query = query;
	}

	private Boolean isDNMode(){
		ListDatabaseDocument outXML = ListDatabaseDocument.Factory.newInstance();
		outXML.addNewListDatabase();
		ListDatabaseResponseDocument inXML = null;
		logger.info("Invoking ...");
		try{
			inXML = admin.listDatabase(outXML);
		}catch(Exception xfe){
			exitWithError("Error invoking service", xfe);
		}
		if(!inXML.getListDatabaseResponse().isSetIsDNMode())return null;
		return inXML.getListDatabaseResponse().getIsDNMode();
	} 

	private void exitWithError(String msg, Exception cause){
		if(cause!=null)logger.error(msg,cause);
		else logger.error(msg);
		String ex=cause!=null?" : "+String.valueOf(cause):"";
		System.err.println(msg +ex);
		System.exit(1);
	}

}
