/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************************/
 

package de.fzj.unicore.xuudb;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.security.cert.Certificate;
import java.security.cert.CertificateEncodingException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;

import org.apache.log4j.Logger;
import org.codehaus.xfire.util.Base64;

public class X509Utils {
	
	private static final Logger logger = Configuration.getLogger(X509Utils.class);
	
	private static final String X509BEGIN_TOKEN = "-----BEGIN CERTIFICATE-----";
	
	private static final String X509END_TOKEN = "-----END CERTIFICATE-----";

	/**
	 * reads a .pem file  (x509 certificate with BASE64 encoding)
	 * it filters all information that might be before the -----BEGIN CERTIFICATE-----
	 * @param pemFile
	 * @return a string which contains the BASE64 encoded cert
	 * @throws FileNotFoundException
	 * @throws link IllegalArgumentException
	 */
	public static String getStringFromPEMFile(File pemFile) throws FileNotFoundException, IllegalArgumentException{
		BufferedReader br = new BufferedReader(new FileReader(pemFile));
		try {
			while (!br.readLine().equalsIgnoreCase(X509BEGIN_TOKEN))
				;
		} catch (Exception e) {
			String msg="Cannot find '"+X509BEGIN_TOKEN+"' in file: "+ pemFile.getAbsolutePath();
			logger.error(msg);
			logger.debug(msg,e);
			throw new IllegalArgumentException(msg);
		}
		StringBuilder cert = new StringBuilder();
		String line="";
		try {
			while (!(line = br.readLine()).equalsIgnoreCase(X509END_TOKEN))
				cert.append (line);
		} catch (Exception e) {
			String msg="Cannot find '"+X509END_TOKEN+"' in file: " + pemFile.getAbsolutePath();
			logger.error(msg);
			logger.debug(msg,e);
			throw new IllegalArgumentException(msg);
		}
		try {
			br.close();
		}
		catch(IOException ex) {}
		return cert.toString();
	}
	
	/**
	 * gets a base64 encoded certificate (linebreaks included)
	 * @param pemstr
	 * @return instance of X509Certificate from pemstr
	 * @throws CertificateException
	 */
	public static X509Certificate getX509FromPEMString(String pemstr) throws CertificateException {
		if( pemstr == null )
			return null;
		String work = pemstr.replace(X509BEGIN_TOKEN, "").replace(X509END_TOKEN, "").replace("\n", "");
		byte[] certbuf = Base64.decode(work);
		CertificateFactory cf = CertificateFactory.getInstance("X.509");
		X509Certificate x509cert=null;
		x509cert = (X509Certificate) cf.generateCertificate(new ByteArrayInputStream(certbuf));
		return x509cert;
	}

	public static String getPEMStringFromX509(Certificate x509) throws CertificateEncodingException {
		String ret;
		byte[] coded = x509.getEncoded();
		byte[] base64 = Base64.encode(coded).getBytes();
		ret=new String(base64);
		return ret;
	}
	
}