package de.fzj.unicore.xuudb.server;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.security.cert.X509Certificate;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.TimeUnit;

import javax.security.auth.x500.X500Principal;

import org.apache.log4j.Logger;
import org.codehaus.xfire.MessageContext;
import org.codehaus.xfire.handler.AbstractHandler;
import org.codehaus.xfire.transport.http.XFireServletController;

/**
 * use an ACL file to limit access
 * @author schuller
 *
 */
public class ACLHandler extends AbstractHandler {

	private static final Logger logger=Logger.getLogger(ACLHandler.class);

	private final File aclFile;
	private final FileWatcher watchDog;
	private final boolean active;
	private final Set<String>acceptedDNs=new HashSet<String>();

	public ACLHandler()throws IOException{
		this(new File("conf","xuudb.acl"));
	}


	public ACLHandler(File aclFile)throws IOException{
		this.aclFile=aclFile;
		if(!aclFile.exists()){
			logger.warn("ACL not active: file <"+aclFile+"> does not exist");
			active=false;
			watchDog=null;
			return;
		}
		else{
			active=true;
			logger.info("XUUDB using ACL file "+aclFile);
			readACL();
			watchDog=new FileWatcher(aclFile, new Runnable(){
				public void run(){
					readACL();
				}
			});
			watchDog.schedule(3000,TimeUnit.MILLISECONDS);
		}
	}

	protected void readACL(){
		synchronized(acceptedDNs){
			BufferedReader br=null;
			try{
				br=new BufferedReader(new FileReader(aclFile));
				String theLine;
				acceptedDNs.clear();
				while(true){
					theLine=br.readLine();
					if(theLine==null)break;
					String line=theLine.trim();
					if(line.startsWith("#"))continue;
					if(!line.trim().equals("")){
						try{
							X500Principal p=new X500Principal(line);
							acceptedDNs.add(p.getName());
							logger.info("Allowing admin access for <"+line+">");	
						}catch(Exception ex){
							logger.warn("Invalid entry <"+line+">",ex);
						}
					}
				}
			}catch(Exception ex){
				logger.fatal("ACL file read error!",ex);
			}
			finally{
				try{
					if(br!=null)br.close();
				}catch(IOException ioex){}
			}
		}
	}

	public void invoke(MessageContext context) throws Exception {
		String userName="anonymous";
		try{
			X509Certificate[] certPath = (X509Certificate[]) XFireServletController.getRequest().getAttribute("javax.servlet.request.X509Certificate");
			if (certPath != null){
				X509Certificate userCert=certPath[0];
				userName=userCert.getSubjectX500Principal().getName();
			}
		}catch(Exception ex){
			logger.error("Can't get user name from request. no ssl?",ex);
		}
		logger.info("Admin access from "+userName);

		if(!active)return;

		checkAccess(userName);
	}

	protected void checkAccess(String userName)throws Exception{
		synchronized (acceptedDNs) {
			if(!acceptedDNs.contains(userName)){
				String msg="Admin access denied!\n\nTo allow access for this " +
						"certificate, the distinuguished name \n\"" +userName+
						"\nneeds to be entered into the ACL file."
						+"\nPlease check the XUUDB's ACL file!\n\n" ;
				throw new Exception(msg);
			}
		}
	}

}
