/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 ********************************************************************************/

package de.fzj.unicore.xuudb.server;


import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.codehaus.xfire.transport.http.XFireServlet;
import org.mortbay.jetty.Connector;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.SessionIdManager;
import org.mortbay.jetty.bio.SocketConnector;
import org.mortbay.jetty.security.SslSocketConnector;
import org.mortbay.jetty.servlet.Context;
import org.mortbay.jetty.servlet.HashSessionIdManager;

import de.fzj.unicore.xuudb.Configuration;
import de.fzj.unicore.xuudb.XUUDBInit;

/**
 * a Jetty server hosting a single xfire servlet
 * 
 * @author schuller
 */
public class JettyServer {
	
	private static final Logger logger=Logger.getLogger(JettyServer.class.getName());
	
	private Server theServer;
	
	private int port;
	private String host;
	
	private String keystore, truststore;
	private String keystorePassword, truststorePassword, keystoreType, truststoreType;
	
	
	public JettyServer(String host,int port)  throws Exception {
		this.host=host;
		this.port=port;
		initServer(false);
	}
	public JettyServer(String host,int port, String keystore, String keystoreType, String keystorePassword)  throws Exception {
		this(host, port, keystore, keystoreType, keystorePassword, keystore, keystoreType, keystorePassword);
	}
	public JettyServer(String host, int port, String keystore, String keystoreType, String keystorePassword, String truststore, String truststoreType, String truststorePassword)  throws Exception {
		this.port=port;
		this.host=host;
		if( Boolean.valueOf(XUUDBInit.getConfig().getProperty(Configuration.USESSL)) 
				&& truststore == null
				&& keystoreType!=null
				&& keystoreType.equalsIgnoreCase("PKCS12") 
				) {
			logger.info("You're using SSL _and_ your truststore is the same file as your keystore _and_ the keystoretype is PKCS12. Make sure you have the correct trusted certificate entries within this file.");
		}

		this.keystore = keystore;
		this.truststore = (truststore==null)?keystore:truststore;
		this.keystoreType =(keystoreType==null)? "JKS":keystoreType;
		this.truststoreType = (truststoreType==null)?"JKS":truststoreType;
		this.keystorePassword = keystorePassword;
		this.truststorePassword = (truststorePassword==null)?keystorePassword:truststorePassword;
		initServer(true);
	}
	
	
	public void start() throws Exception {
		theServer.start();
	}
	
	public void stop() throws Exception {
		theServer.stop();
	}
	
	private void initServer(boolean ssl) throws Exception {
		logger.log(Level.INFO,"Starting Jetty on port :"+port);
		theServer=new Server();
		SessionIdManager sm = new HashSessionIdManager(new java.util.Random());
		theServer.setSessionIdManager(sm);
		Connector connector=ssl?createSecureConnector():createConnector();
		theServer.addConnector(connector);
		configureConnector(connector);
		addServlets();
	}
	
	private Connector createConnector() {
		logger.info("Creating plain HTTP connector.");
		Connector connector= new SocketConnector();
		configureConnector(connector);
		return connector;
	}
	
	private Connector createSecureConnector() {
		logger.info("Creating SSL connector.");
		SslSocketConnector ssl=new SslSocketConnector();
		//configure ssl listener
		ssl.setNeedClientAuth(true);
		ssl.setKeystore(this.keystore);
		ssl.setPassword(this.keystorePassword);
		ssl.setKeyPassword(this.keystorePassword);
		ssl.setKeystoreType(this.keystoreType);
		ssl.setTruststore(this.truststore);
		ssl.setTruststoreType(this.truststoreType);
		ssl.setTrustPassword(this.truststorePassword);
		configureConnector(ssl);
		//fix for IBM JDK where default protocol "TLS" does not work
		String vm=System.getProperty("java.vm.vendor");
		if(vm!=null && vm.trim().startsWith("IBM")){
			ssl.setProtocol("SSL_TLS");
			logger.info("For IBM JDK: Setting SSL protocol to '"+ssl.getProtocol()+"'");
		}
		return ssl;
	}
	
	
	//TODO: could do some more config here (max threads and the like)
	private void configureConnector(Connector http){
		http.setPort(port);
		http.setHost(host);
		
//		http.setMaxIdleTime(1000);
//		http.setLowResourceMaxIdleTime(100);
	}
	
	
	private void addServlets()throws Exception {
		logger.log(Level.INFO,"Adding XFire servlet.");// on "+Kernel.getKernel().getProperty(Kernel.WSRF_SERVLETPATH)+"/*");
		Context root=new Context(theServer, "/", Context.SESSIONS);
		root.addServlet(XFireServlet.class,"/*");
	}

	/**
	 * get the Jetty server
	 * @return server
	 */
	public Server getServer(){
		return theServer;
	}
}
