package de.fzj.unicore.xuudb.server;

import java.security.cert.X509Certificate;

import javax.security.auth.x500.X500Principal;

import org.apache.log4j.Logger;

import de.fzj.unicore.xuudb.Configuration;
import de.fzj.unicore.xuudb.X509Utils;

/**
 * 
 * this class is a wrapper for managing tokens in pem or dn format.
 * 
 * In DN mode, the DN used as token is canonicalized according to RFC2253
 * @see {X500Principal}
 * @author tweddell
 * @author schuller
 * @since 1.0.1
 */

public class SecurityToken {

	private static final Logger logger = Configuration.getLogger(SecurityToken.class);
	
	private final String token;
	private final boolean dnMode;
	
	/**
	 * create a new security token
	 * 
	 * @param token - the raw token
	 * @param dnMode - if the DB is running in dnMode
	 * @throws Exception
	 */
	public SecurityToken(String token, boolean dnMode) throws Exception {
    	this.dnMode=dnMode;
		this.token=buildToken(token);
	}
	
	
	protected String buildToken(String token)throws Exception{
		String t=null;
		//check which type of xuudb database we have: use "normal" as default
		if(dnMode) {
			logger.debug("Creating Security token for dn database");
			try {
				t = new X500Principal(token).getName(X500Principal.RFC2253);
			} catch (Exception e) {
				if(logger.isDebugEnabled()){
					logger.debug("Cannot create an instance of X500Name from: " + token
							+"\nTrying to get an instance of X509Certificate and then extract the DN...");
				}
				try {
					X509Certificate cert = X509Utils.getX509FromPEMString(token);
					t = cert.getSubjectX500Principal().getName(X500Principal.RFC2253);		
				} catch (Exception ce) {
					logger.error("Token is not a DN or an x509 cert. Cannot create security token from: " + token);
					throw new IllegalArgumentException("Token is not a DN or x509 cert.");
				}
			}
		}
		else{
			logger.debug("Creating Security token for normal database");			
			try {
				X509Utils.getX509FromPEMString(token);
				t = token.replace("\n", "");
			} catch (Exception e) {
				String msg="Sorry, but that's not a pem encoded X509Certificate: " + token+
				"\nDatabase mode is set to 'normal'. Only full certificates can be used.";
				logger.info(msg);
				throw new IllegalArgumentException(msg);
			}
		}
		return t;
	}

	
	@Override
	public String toString() {
		return this.token;
	}
	
	@Override
	public boolean equals(Object obj) {
		if(! (obj instanceof SecurityToken)) {
			return false;
		}
		return this.token.equalsIgnoreCase(obj.toString());
	}
	
	public static String getDN(String token, boolean isDNMode)throws Exception{
		if(isDNMode){
			return new X500Principal(token).getName();
		}
		return X509Utils.getX509FromPEMString(token).getSubjectX500Principal().getName();
	}

	public int hashCode() {
		//equal tokens (see equals() method) should have equal hashcodes
		return token.toLowerCase().hashCode()^0x36593265;
	}
}
