/*********************************************************************************
 * Copyright (c) 2008 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************************/


package de.fzj.unicore.xuudb.server;

import java.util.ArrayList;
import java.util.List;

import javax.security.auth.x500.X500Principal;

import org.apache.log4j.Logger;

import de.fzJuelich.unicore.xuudb.ImportDatabaseDocument;
import de.fzJuelich.unicore.xuudb.LoginDataType;
import de.fzj.unicore.xuudb.Configuration;
import de.fzj.unicore.xuudb.XUUDBInit;
import de.fzj.unicore.xuudb.server.db.IStorage;

/**
 * Storage logic<br/>
 * Delegates actual storage operations to
 * an instance of {@link IStorage} that is injected by
 * {@link #setDatabase(IStorage)}
 * 
 * @author schuller
 */
public class XUUDB {

	private IStorage database;

	private static final Logger logger = Configuration.getLogger(XUUDB.class);

	public void setDatabase(IStorage db) throws Exception {
		database=db;
	}

	public XUUDB(){
		logger.info("Starting XUUDB");
		String v=XUUDB.class.getPackage().getImplementationVersion();
		if(v!=null){
			logger.info("Server version: "+v);
		}
		String dbtype=XUUDBInit.getConfig().getProperty(Configuration.XUUDBTYPE,"normal");
		logger.info("Running in "+dbtype.toUpperCase()+" mode.");
	}

	public String add(LoginDataType addRequest) {
		logger.info("Adding into UUDB");
		try { 
			//first need to check if an entry exists
			String gcid=addRequest.getGcID();
			String token=addRequest.getToken();
			List<LoginDataType> existingEntries=findEntries(gcid, token);
			if(existingEntries.size()==0){
				database.add(addRequest);
			}
			else{
				for(LoginDataType existingEntry: existingEntries){
					String newXlogin=addRequest.getXlogin();
					Xlogin xlogin=new Xlogin(existingEntry.getXlogin());
					boolean added=xlogin.addLogin(newXlogin);
					if(added){
						logger.info("Adding new xlogin <"+newXlogin+"> to existing entry.");
						existingEntry.setXlogin(xlogin.getEncoded());
						database.update(gcid, token, existingEntry);
					}
					else{
						logger.info("Xlogin <"+newXlogin+"> already exists in entry.");
					}
				}
			}
		} catch (Exception e) {
			logger.error("Error while adding.. STOP",e);
			return e.getMessage();
		}

		return "OK.";
	}

	/**
	 * check if an entry exists for the given gcid and token
	 * @param gcid - the GCID. Wildcard "*" or just <code>null</code> is allowed
	 * @param token - the token (DN or X509 pem file)
	 * @return a List of {@link LoginDataType} which is empty if nothing is found
	 */
	protected List<LoginDataType> findEntries(String gcid, String token){
		List<LoginDataType>result=new ArrayList<LoginDataType>();
		List<String>gcids=null;
		if(gcid==null || "*".equals(gcid)){
			gcids=getGCIDs();
		}
		else{
			gcids=new ArrayList<String>();
			gcids.add(gcid);
		}
		
		for(String gc: gcids){
			try{
				boolean isDN=false;
				X500Principal dn=null;
				if(token!=null){
					try{
						dn=new X500Principal(token);
						isDN=true;
					}
					catch(Exception ignored){}
				}
				LoginDataType entry=isDN?database.checkDN(gc, dn.getName()):database.checkToken(gc, token);
				if(entry!=null && entry.getXlogin()!=null){
					result.add(entry);
				}
			}catch(Exception ex){
				logger.error("Error finding entry.", ex);
			}
		}
		return result;
	}


	public String remove(LoginDataType removeRequest) {
		logger.info("Removing from UUDB");
		StringBuilder ret=new StringBuilder();
		try { 
			//first need to check if an entry exists
			String gcid=removeRequest.getGcID();
			String token=removeRequest.getToken();
			List<LoginDataType> existingEntries=findEntries(gcid, token);
			if(existingEntries.size()==0)return removeByTemplate(removeRequest);
			for(LoginDataType existingEntry: existingEntries){
				//remove/edit
				Xlogin xlogin=new Xlogin(existingEntry.getXlogin());
				String removeXlogins=removeRequest.getXlogin();
				if(removeXlogins==null || xlogin.getNumberOfLogins()==1){
					logger.info("Removing entry.");
					ret.append(database.remove(existingEntry)).append("\n");
				}
				else{
					//update
					Xlogin toRemove=new Xlogin(removeXlogins);
					for(String login: toRemove){
						boolean removed=xlogin.removeLogin(login);
						if(!removed){
							return "Xlogin "+login+" not found in entry "+xlogin+".Operation aborted.";
						}
					}
					if(xlogin.getNumberOfLogins()==0){
						logger.info("Removing entry.");
						ret.append(database.remove(existingEntry)).append("\n");
					}
					else{
						logger.info("Updating xlogin to <"+xlogin+">.");
						existingEntry.setXlogin(xlogin.getEncoded());
						ret.append(database.update(gcid, token, existingEntry)).append("\n");
					}
				}
			}
		} catch (Exception e) {
			logger.error("Error while removing... STOP",e);
			ret.append("Server error occurred, please refer to the server log file for details.").append("\n");
		}
		return ret.toString();
	}

	protected String removeByTemplate(LoginDataType removeRequest){
		String ret;
		try{
			ret=database.remove(removeRequest);
		}catch(Exception e){
			logger.error("Error while removing... STOP",e);
			ret="A server error occurred while removing. Please refer to the server log for details.";
		}
		return ret;
	}

	public String import_csv(ImportDatabaseDocument idd) {
		return database.import_csv(idd);
	}

	public LoginDataType[] listDB(LoginDataType query) {
		return database.listDB(query);
	}

	/**
	 * reads login info from the database
	 * 
	 * @param gcid
	 * @param certinpem
	 * @return login info
	 */
	public LoginDataType checkCert(String gcid, String token) {
		return database.checkToken(gcid, token);
	}

	/**
	 * reads login info from the database by checking DN using X500Principal.getName()
	 * 
	 * @param gcid
	 * @param certinpem
	 * @return login info
	 */
	public LoginDataType checkDN(String gcid, String dn) {
		return database.checkDN(gcid, dn);
	}


	public String update(String gcid, String token, LoginDataType login) {
		logger.info("updateDB Request");
		return database.update(gcid, token, login);
	}


	public List<String>getGCIDs() {
		logger.info("getGCID Request");
		return database.listGCIDs();
	}

	public void shutdown(){
		try{
			database.shutdown();
		}catch(Exception e){
			logger.warn("Problem shutting down.",e);
		}
	}
}
