/*********************************************************************************
 * Copyright (c) 2006 Forschungszentrum Juelich GmbH 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * (1) Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the disclaimer at the end. Redistributions in
 * binary form must reproduce the above copyright notice, this list of
 * conditions and the following disclaimer in the documentation and/or other
 * materials provided with the distribution.
 * 
 * (2) Neither the name of Forschungszentrum Juelich GmbH nor the names of its 
 * contributors may be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * 
 * DISCLAIMER
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************************/


package de.fzj.unicore.xuudb.server.db;

import java.security.cert.CertificateException;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.security.auth.x500.X500Principal;

import org.apache.log4j.Logger;

import de.fzJuelich.unicore.xuudb.DatabaseType;
import de.fzJuelich.unicore.xuudb.ImportDatabaseDocument;
import de.fzJuelich.unicore.xuudb.LoginDataType;
import de.fzj.unicore.xuudb.Configuration;
import de.fzj.unicore.xuudb.IShutdownable;
import de.fzj.unicore.xuudb.ShutdownHook;
import de.fzj.unicore.xuudb.X509Utils;
import de.fzj.unicore.xuudb.XUUDBInit;
import de.fzj.unicore.xuudb.server.SecurityToken;

/**
 * the storage backend of the XUUDB
 */
public class Database implements IShutdownable, IStorage {

	private boolean isDNMode;

	private final String SQLAddTable = 
		"CREATE TABLE uudb ( " +
		"gcid VARCHAR(255), " +
		"xlogin VARCHAR(255), " +
		"token VARCHAR(8192), " +
		"role VARCHAR(255), " +
		"projects VARCHAR(255), " +
		"UNIQUE(gcid,token))";


	private PreparedStatement addCertSm;
	private PreparedStatement checkAuthorisation;
	private PreparedStatement listGCIDs;

	private static final Logger logger = Configuration.getLogger(Database.class);

	private Connection conn;

	public Database(String dbFileName) throws Exception {
		logger.info("Loading Database from file: " + dbFileName);
		// Load the HSQL Database Engine JDBC driver
		// hsqldb.jar should be in the class path or made part of the current jar
		try {
			logger.debug("Loading driver: org.hsqldb.jdbcDriver");
			Class.forName("org.hsqldb.jdbcDriver" );
		} catch (Exception e) {
			logger.error("Failed to load JDBC driver", e);
			throw e;
		}
		// connect to the database.   This will load the db files and start the
		// database if it is not alread running.
		// db_file_name_prefix is used to open or create files that hold the state
		// of the db.
		// It can contain directory names relative to the
		// current working directory
		try {
			logger.info("Connecting to the DB: jdbc:hsqldb:file:" +
					dbFileName);
			conn = DriverManager.getConnection("jdbc:hsqldb:file:"+
					dbFileName, "sa", "");
		} catch (SQLException e) {
			logger.error("Cannot create database connection: " + e.getMessage());
			throw e;
		}
		// register in shutdownhook
		ShutdownHook.register(this);
		try {
			// make an empty table
			// by declaring the gciid column IDENTITY, the db will automatically
			// generate unique values for new rows- useful for row keys
			Statement st = null;
			st = conn.createStatement();    // statements
			int i = st.executeUpdate(SQLAddTable);    // run the query
			if (i == -1) {
				logger.error("db error : " + SQLAddTable);
			}
			st.close();
		} catch (SQLException ex2) {
			// ignore ex
			// second time we run program
			// should throw execption since table
			// already there
			// this will have no effect on the db
			logger.trace("While updating the DB an exception was thrown: " + ex2.getMessage());
		}		

		try {
			createPreparedStatements();
		} catch (SQLException ex) {
			logger.error("Error while creating Prepared Statements: " + ex.getMessage());
			throw ex;
		}

		try{
			String dbtype=XUUDBInit.getConfig().getProperty(Configuration.XUUDBTYPE,"normal");
			isDNMode="dn".equalsIgnoreCase(dbtype);
		}catch(Exception e){
			logger.error("config problem?!",e);
		}

		logger.info("Database HSQLDB Started");
	}

	private void createPreparedStatements() throws SQLException {
		//creating prepared statements
		addCertSm = conn.prepareStatement("INSERT INTO uudb (gcid,xlogin,token,role,projects) VALUES (?,?,?,?,?)");
		checkAuthorisation = conn.prepareStatement("SELECT * FROM uudb WHERE gcid = ? AND token = ?");
		listGCIDs= conn.prepareStatement("SELECT gcid FROM uudb");
	}

	public void shutdown() throws Exception {
		synchronized(conn){
			if(conn.isClosed())return;
			logger.info("Disconnecting from Database");
			Statement st = conn.createStatement();
			// db writes out to files and performs clean shuts down
			// otherwise there will be an unclean shutdown
			// when program ends
			st.execute("SHUTDOWN");
			st.close();
			conn.close();    // if there are no other open connections
		}
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#add(de.fzJuelich.unicore.xuudb.LoginDataType)
	 */
	public String add(LoginDataType addRequest) {
		//    	INSERT INTO uudb (gcid,xlogin,token,role,projects) VALUES (?,?,?,?,?)
		synchronized (conn) {

			SecurityToken addtok; 
			try { // create SecurityToken
				addtok = new SecurityToken(addRequest.getToken(),isDNMode); 
			} catch (Exception e) {
				logger.error("Error while adding.. STOP",e);
				return e.getMessage();
			}

			try {
				addCertSm.setString(1, addRequest.getGcID());
				addCertSm.setString(2, addRequest.getXlogin());
				addCertSm.setString(3, addtok.toString());
				addCertSm.setString(4, addRequest.getRole());
				addCertSm.setString(5, addRequest.getProjects());
			} catch (SQLException e) {
				logger.error("Error creating prepared statement: " + e.getMessage());
				return "Error creating prepared statement: " + e.getMessage();
			}
			logger.info(addCertSm.toString());

			try {
				addCertSm.executeUpdate();
				flushDB();
			} catch (SQLException e) {
				logger.error("Error while inserting into database: " + e.getMessage());
				return "Error while inserting into database: " + e.getMessage();
			}
			return "OK.";
		}
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#import_csv(de.fzJuelich.unicore.xuudb.ImportDatabaseDocument)
	 */
	public String import_csv(ImportDatabaseDocument idd) {
		synchronized(conn){
			DatabaseType db = idd.getImportDatabase();
			LoginDataType[] logins = db.getDatabaseArray();
			int occuredErrors=0;
			boolean clear=false;
			if(idd.getImportDatabase().isSetClean()){
				clear=idd.getImportDatabase().getClean();
			}
			if(clear){
				logger.info("Clearing database before import.");
				//remove all the entries
				this.remove(null);
			}
			for(int i = 0; i<logins.length; i++) {
				String erg = this.add(logins[i]);
				if( !erg.equals("OK.")) occuredErrors++;
			}

			return (logins.length-occuredErrors) + "   certificates from   "  + logins.length + 
			"   were imported into the XUUDB.";
		}
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#listDB(de.fzJuelich.unicore.xuudb.LoginDataType)
	 */
	public LoginDataType[] listDB(LoginDataType query) {
		synchronized(conn){
			LoginDataType[] ret;
			Statement listDbSm = null;
			try {
				listDbSm = conn.createStatement();
			} catch (SQLException e) {
				logger.error("Error creating statement for db query : " + e.getMessage());
				return null;
			}
			String sqlSelect;
			try {
				sqlSelect = "SELECT * FROM uudb" + buildWhereFromUlogin(query);
			} catch (Exception e1) {
				logger.error("Cannot build Where Statement, because of incorrect token. Mistyped DN or wrong cert.");
				return null;
			}
			logger.info(sqlSelect);
			try {
				ResultSet rs = listDbSm.executeQuery(sqlSelect);
				ArrayList<LoginDataType> data = new ArrayList<LoginDataType>();
				while(rs.next()) {
					LoginDataType adder = LoginDataType.Factory.newInstance();
					adder.setGcID(rs.getString("gcid"));
					adder.setXlogin(rs.getString("xlogin"));
					adder.setToken(rs.getString("token"));
					adder.setRole(rs.getString("role"));
					adder.setProjects(rs.getString("projects"));
					data.add(adder);
				}
				//build return value
				ret = new LoginDataType[data.size()];
				Iterator<LoginDataType> it = data.iterator();
				for(int i=0;it.hasNext();++i) {
					ret[i] = (LoginDataType) it.next();
				}
				listDbSm.close();
			} catch (SQLException e) {
				logger.error("Error while querying the database: " + e.getMessage());
				return null;
			}
			return ret;
		}
	}

	private String buildWhereFromUlogin(LoginDataType query) throws Exception {
		String ret = "";
		if (query != null) {
			ret += " WHERE ";
			String adder = null;
			boolean doAnd = false;
			if ((adder = query.getGcID()) != null) {
				ret += "gcid='" + adder + "'";
				doAnd = true;
			}
			if ((adder = query.getXlogin()) != null) {
				if (doAnd)
					ret += " AND ";
				ret += "xlogin='" + adder + "'";
				doAnd = true;
			}
			if ((adder = query.getToken()) != null) {
				if (doAnd)
					ret += " AND ";
				adder = new SecurityToken(adder,isDNMode).toString();
				ret += "token='" + adder + "'";
				doAnd = true;
			}
			if ((adder = query.getRole()) != null) {
				if (doAnd)
					ret += " AND ";
				ret += "role='" + adder + "'";
				doAnd = true;
			}
			if ((adder = query.getProjects()) != null) {
				if (doAnd)
					ret += " AND ";
				ret += "projects LIKE '%" + adder + "%'";
				doAnd = true;
			}
		}
		logger.trace("Built statement: " + ret);
		if(ret.equalsIgnoreCase(" WHERE ")) {
			logger.info("Dropping WHERE Expression for query. listing all.");
			return "";
		}
		return ret;
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#remove(de.fzJuelich.unicore.xuudb.LoginDataType)
	 */
	public String remove(LoginDataType query) {
		synchronized(conn){
			//DELETE FROM uudb WHERE ....
			Statement listDbSm = null;
			try {
				listDbSm = conn.createStatement();
			} catch (SQLException e) {
				logger.error("Error creating statement for db query : " + e.getMessage());
				return "Error creating statement for db query : " + e.getMessage();
			}
			String sqlDelete;
			try {
				sqlDelete = "DELETE FROM uudb" + buildWhereFromUlogin(query);
			} catch (Exception e1) {
				logger.error("Cannot build Where Statement, because of incorrect token. Mistyped DN or wrong cert.");
				return "Cannot build Where Statement, because of incorrect token. Check your typed DN or given cert.";
			}
			logger.info(sqlDelete);
			String ret = "";
			try {
				int erg = listDbSm.executeUpdate(sqlDelete);
				ret = "Removed lines: " + erg;
				listDbSm.close();
				flushDB();
			} catch (SQLException e) {
				logger.error("Error while deleting from the database: " + e.getMessage());
				return "Error while deleting from the database: " + e.getMessage();
			}
			return ret;
		}
	}

	public String getNameOfService() {
		return "XUUDB Database using HSQLDB";
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#checkCert(java.lang.String, java.lang.String)
	 */
	public LoginDataType checkToken(String gcid, String token) {
		if(token==null)return null;
		synchronized(conn){
			LoginDataType ret = LoginDataType.Factory.newInstance();
			if(logger.isDebugEnabled()){
				String msg="Check for matching token "+(gcid!=null ? "in gcid "+gcid : "" );
				logger.info(msg);
			}
			token = token.replace("\n", "");
			try {
				String vergleich = new SecurityToken(token,isDNMode).toString();
				checkAuthorisation.setString(1, gcid);
				checkAuthorisation.setString(2, vergleich);
			} catch (Exception e) {
				logger.error("Error creating prepared statement: " + e.getMessage());
				return ret;
			}
			if(logger.isDebugEnabled())logger.debug(checkAuthorisation.toString());
			ResultSet rs = null;
			try {
				rs = checkAuthorisation.executeQuery();
			} catch (SQLException e) {
				logger.error("Error while querying the database: " + e.getMessage());
				return ret;
			}
			try {
				if(rs.next()) {
					ret.setGcID(rs.getString("gcid"));
					ret.setXlogin(rs.getString("xlogin"));
					ret.setToken(rs.getString("token"));
					ret.setRole(rs.getString("role"));
					ret.setProjects(rs.getString("projects"));
					logger.info("XUUDB has a matching entry: xlogin = " + ret.getXlogin());
				} else {
					String dn = null;
					try{
						dn=new X500Principal(token).getName();
					}catch(Exception ignored){}
					if(dn==null){
						try {
							dn = X509Utils.getX509FromPEMString(token).getSubjectDN().getName();
						} catch (CertificateException e) {
							logger.info("Cannot create DN from cert I got from XNJS.", e);
						}
					}
					logger.debug("No matching entry found in UUDB for DN: " + dn);
				}
				rs.close();
			} catch (SQLException e) {
				logger.error("Error while getting the result set: " + e.getMessage());
			}		

			return ret;
		}
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#checkDN(java.lang.String, java.lang.String)
	 */
	public LoginDataType checkDN(String gcid, String dn) {
		//if in dn mode, we can delegate
		if(isDNMode){
			return checkToken(gcid,dn);
		}
		//else we have to check first whether a matching cert exists,
		//and then we can get the attributes for that cert
		synchronized(conn){
			LoginDataType query=LoginDataType.Factory.newInstance();
			query.setGcID(gcid);
			LoginDataType[] all=listDB(query);
			for(LoginDataType r: all) {
					try{
						String toCheck=new X500Principal(dn).getName();
						String fromDB=SecurityToken.getDN(r.getToken(), isDNMode);
						logger.debug("Checking <"+toCheck+"> vs XUUDB entry <"+fromDB+">");
						//logger.finer(toCheck)
						if(toCheck.equals(fromDB)){
							logger.debug("XUUDB has a cert with matching DN --> returning attributes now");
							return r;
						}
					}catch(Exception e){
						logger.error("",e);
					}
				}
			
			return null;
		}
	}

	/* (non-Javadoc)
	 * @see de.fzj.unicore.xuudb.server.db.IStorage#update(java.lang.String, java.lang.String, de.fzJuelich.unicore.xuudb.LoginDataType)
	 */
	public String update(String gcid, String token, LoginDataType login) {
		logger.info("updateDB Request");
		synchronized(conn){
			Statement updateDbSm = null;
			try {
				updateDbSm = conn.createStatement();
			} catch (SQLException e) {
				logger.error("Error creating statement for db update : " + e.getMessage());
				return null;
			}
			// token might be in pem or dn => get correct formatted token from Securitytoken
			try {
				token = new SecurityToken(token,isDNMode).toString();
			} catch (Exception e1) {
				logger.error("Cannot update database, because I cannot create SecurityToken.");
				return "Cannot update database, because I cannot create SecurityToken.";
			}
			if(login.getToken()!=null){
				//need correct token in DN mode
				try {
					String newToken = new SecurityToken(login.getToken(),isDNMode).toString();
					login.setToken(newToken);
				} catch (Exception ex) {
					String msg="Cannot update database, because I cannot create SecurityToken.";
					logger.error(msg,ex);
					return msg;
				}
			}
			String sqlUpdate = "UPDATE uudb " + buildSetFromUlogin(login);
			
			if(gcid.equals("*")){
				sqlUpdate += " WHERE token='" + token +"'";
			}
			else{
				sqlUpdate += " WHERE gcid='" + gcid + "' AND token='" + token +"'";
			}
			logger.info(sqlUpdate);
			int anz=0;
			try {
				anz = updateDbSm.executeUpdate(sqlUpdate);
				updateDbSm.close();
				flushDB();
			} catch (SQLException e) {
				logger.error("Error while querying the database: " + e.getMessage());
				return null;
			}

			return "Updated   " + anz + "   rows." ;
		}
	}

	public List<String>listGCIDs(){
		List<String>result=new ArrayList<String>();
		try{
			synchronized(conn){
				ResultSet rs=listGCIDs.executeQuery();
				while(rs.next()){
					String r=rs.getString(1);
					result.add(r);
				}
				rs.close();
			}
		}
		catch(SQLException ex){
			logger.error("", ex);
		}
		return result;
	}
	
	private String buildSetFromUlogin(LoginDataType login) {
		logger.trace("Building SET statement..");
		String ret = "";
		if (login != null) {
			ret += " SET  ";
			String adder = null;
			boolean doComma = false;
			if(!"*".equals(login.getGcID())){
				if ((adder = login.getGcID()) != null) {
					ret += "gcid='" + adder + "'";
					doComma = true;
				}
			}
			if ((adder = login.getXlogin()) != null) {
				if (doComma)
					ret += ", ";
				ret += "xlogin='" + adder + "'";
				doComma = true;
			}
			if ((adder = login.getToken()) != null) {
				if (doComma)
					ret += ", ";
				ret += "token='" + adder + "'";
				doComma = true;
			}
			if ((adder = login.getRole()) != null) {
				if (doComma)
					ret += ", ";
				ret += "role='" + adder + "'";
				doComma = true;
			}
			if ((adder = login.getProjects()) != null) {
				if (doComma)
					ret += ", ";
				ret += "projects='" + adder + "'";
				doComma = true;
			}
		}
		return ret;
	}

	/**
	 *checkpoint the hsqldb database. 
	 *this should be called after write access to the uudb
	 */
	private void flushDB(){
		Statement s=null;
		try{
			s=conn.createStatement();
			s.execute("CHECKPOINT");
			s.close();
		}catch(Exception e){
			logger.error("Error checkpointing database to disk.",e);
		}
		finally{
			try{
				s.close();
			}catch(Exception e){}
		}
	}

	public boolean isDNMode(){
		return isDNMode;
	}

}
