package de.fzj.unicore.xuudb.server;

import java.io.File;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.TreeMap;

import junit.framework.TestCase;
import de.fzJuelich.unicore.xuudb.LoginDataType;
import de.fzj.unicore.xuudb.Configuration;
import de.fzj.unicore.xuudb.X509Utils;
import de.fzj.unicore.xuudb.XUUDBInit;
import de.fzj.unicore.xuudb.server.db.StorageFactory;
import eu.unicore.bugsreporter.annotation.FunctionalTest;
import eu.unicore.bugsreporter.annotation.FunctionalTests;

public class TestXUUDB extends TestCase {
	
	static XUUDB xuudb;
	private static String file;
	private static boolean haveInit=false;
	
	@Override
	protected void setUp()throws Exception{
		if(haveInit){
			LoginDataType removeRequest=LoginDataType.Factory.newInstance();
			xuudb.remove(removeRequest);
			return;
		}
		haveInit=true;
		Properties prop=new Properties();
		prop.setProperty(Configuration.XUUDBTYPE, getMode());
		new XUUDBInit(prop);
		xuudb=new XUUDB();
		file=System.getProperty("java.io.tmpdir")+File.separator+"xuudb_test_"+System.currentTimeMillis();
		xuudb.setDatabase(StorageFactory.getDatabase(file));
	}
	
	protected String getMode(){
		return "normal";
	}
	
	protected boolean isDNMode(){
		return "dn".equals(getMode());
	}	
	
	@FunctionalTests({@FunctionalTest(id="xuudb_remove"), 
			@FunctionalTest(id="xuudb_add"),
			@FunctionalTest(id="xuudb_list"),
			@FunctionalTest(id="xuudb_update")})
	public void testAddRemoveIndividualXlogins()throws Exception{
		
		LoginDataType addRequest=getAddRequest("xlogin1");
		xuudb.add(addRequest);
		
		assertEquals(1,getEntries().length);

		//add a second xlogin
		addRequest=getAddRequest("xlogin2");
		xuudb.add(addRequest);
		
		LoginDataType[]os=getEntries();
		assertEquals(1,os.length);
		LoginDataType x=os[0];
		assertEquals("xlogin1:xlogin2",x.getXlogin());
		
		//add a third xlogin
		addRequest=getAddRequest("xlogin3");
		xuudb.add(addRequest);
		
		os=getEntries();
		assertEquals(1,os.length);
		x=os[0];
		assertEquals("xlogin1:xlogin2:xlogin3",x.getXlogin());
		
		//remove one
		LoginDataType removeRequest=getRemoveRequest("xlogin2");
		xuudb.remove(removeRequest);
		os=getEntries();
		assertEquals(1,os.length);
		x=os[0];
		assertEquals("xlogin1:xlogin3",x.getXlogin());
		//remove 2nd one
		removeRequest=getRemoveRequest("xlogin3");
		xuudb.remove(removeRequest);
		os=getEntries();
		assertEquals(1,os.length);
		x=os[0];
		assertEquals("xlogin1",x.getXlogin());
		
		//remove last one
		removeRequest=getRemoveRequest("xlogin1");
		xuudb.remove(removeRequest);
		os=getEntries();
		assertEquals(0,os.length);
		
	}

	public void testDoNotAddSameXlogin()throws Exception{
		
		LoginDataType addRequest=getAddRequest("xlogin1");
		xuudb.add(addRequest);
		
		assertEquals(1,getEntries().length);

		Xlogin x=new Xlogin(getEntries()[0].getXlogin());
		assertEquals(1, x.getNumberOfLogins());

		addRequest=getAddRequest("xlogin2");
		xuudb.add(addRequest);

		x=new Xlogin(getEntries()[0].getXlogin());
		assertEquals(2, x.getNumberOfLogins());
		
		//add the same xlogin
		addRequest=getAddRequest("xlogin1");
		xuudb.add(addRequest);

		x=new Xlogin(getEntries()[0].getXlogin());
		assertEquals(2, x.getNumberOfLogins());
		
	}

	public void testRemoveFullXlogin()throws Exception{
		Properties prop=new Properties();
		new XUUDBInit(prop);
		xuudb=new XUUDB();
		xuudb.setDatabase(StorageFactory.getDatabase(System.getProperty("java.io.tmpdir")+File.separator+"xuudb_test_"+System.currentTimeMillis()));
		
		LoginDataType addRequest=getAddRequest("xlogin1:xlogin2:xlogin3");
		xuudb.add(addRequest);
		assertEquals(1,getEntries().length);
		
		//remove last one
		LoginDataType removeRequest=getRemoveRequest("xlogin1:xlogin2:xlogin3");
		xuudb.remove(removeRequest);
		LoginDataType[]os=getEntries();
		assertEquals(0,os.length);
		
	}

	public void testRemoveByDN()throws Exception{
		Properties prop=new Properties();
		new XUUDBInit(prop);
		xuudb=new XUUDB();
		xuudb.setDatabase(StorageFactory.getDatabase(System.getProperty("java.io.tmpdir")+File.separator+"xuudb_test_"+System.currentTimeMillis()));
		
		LoginDataType addRequest=getAddRequest("xlogin1:xlogin2:xlogin3");
		xuudb.add(addRequest);
		assertEquals(1,getEntries().length);
		
		//remove last one
		LoginDataType removeRequest=getRemoveByDNRequest();
		xuudb.remove(removeRequest);
		LoginDataType[]os=getEntries();
		assertEquals(0,os.length);
		
	}

	
	public void testBasicCRUD()throws Exception{
		LoginDataType addRequest=getAddRequest("xlogin1");
		xuudb.add(addRequest);
		assertEquals(1,getEntries().length);

		LoginDataType queryResponse=doQuery();
		assertEquals("xlogin1",queryResponse.getXlogin());
		
		List<String>gcids=xuudb.getGCIDs();
		assertNotNull(gcids);
		assertEquals(1,gcids.size());
		assertEquals("foo",gcids.get(0));
		
		
		LoginDataType removeRequest=getRemoveByXloginRequest("xlogin1");
		xuudb.remove(removeRequest);
		assertEquals(0,getEntries().length);

		//removeRequest=getRemoveByXloginRequest("\'; drop table uudb; --");
		//xuudb.remove(removeRequest);

		queryResponse=doQuery();
		assertNull(queryResponse.getXlogin());
		
	
	}

	@FunctionalTest(id="xuudb_performance")
	public void tstPerformance() throws Exception{
		final int ITERATIONS = 10, DIFFERENT_QUERIES = 100, MULTIPLIER = 5;
		final int RECORD_COUNT = DIFFERENT_QUERIES*MULTIPLIER;
		for(int i=0; i<RECORD_COUNT; i++){
			xuudb.add(getAddRequest(i));
		}
		LoginDataType x;
		long time = System.currentTimeMillis();
		for(int j=0; j<ITERATIONS; j++){
			long time0 = System.currentTimeMillis();
			for(int i=0; i<DIFFERENT_QUERIES; i++){
				int no=(i*MULTIPLIER*(j*10+3) + (j^0xffff)) % RECORD_COUNT;
				try {
					x = doQuery("foo" + no);
					assertEquals("user"+no, x.getRole());
					assertEquals("login"+no, x.getXlogin());
					assertEquals("projects"+no, x.getProjects());
				} catch (Exception ex) {
					ex.printStackTrace();
					fail();
				}
			}
			time0 = System.currentTimeMillis() - time0;
			System.out.println("Speed in iteration #"+(j+1)+": " + (1000.0*DIFFERENT_QUERIES/time0) + " queries/s");
		}
		time = System.currentTimeMillis() - time;
		System.out.println("Avarage speed: " + (ITERATIONS*1000.0*DIFFERENT_QUERIES/time) + " queries/s");

		LoginDataType removeRequest=LoginDataType.Factory.newInstance();
		xuudb.remove(removeRequest);
	}

	@FunctionalTest(id="xuudb_scalability")
	public void tstScalability() throws Exception{
		final int ITERATIONS = 200, QUERY_THREADS = 40, CRUD_THREADS = 10;
		final int RECORD_COUNT = 100;
		for(int i=0; i<RECORD_COUNT; i++){
			xuudb.add(getAddRequest(i));
		}
		final Map<Integer, Throwable> errors = Collections.synchronizedMap(new TreeMap<Integer, Throwable>());
		final Thread[] threads = new Thread[QUERY_THREADS + CRUD_THREADS];
		for(int i=0; i<QUERY_THREADS; i++){
			final int copyOfI = i;
			threads[copyOfI]=new Thread(new Runnable() {
				@Override
				public void run() {
					for(int j=0; j<ITERATIONS; j++){
						try {
							LoginDataType queryResponse = doQuery("foo" + copyOfI);
							assertEquals("user" + copyOfI, queryResponse.getRole());
							assertEquals("login" + copyOfI, queryResponse.getXlogin());
							assertEquals("projects" + copyOfI, queryResponse.getProjects());
						} catch (AssertionError ex) {
							errors.put(copyOfI, ex);
							return;
						} catch (Exception ex) {
							errors.put(copyOfI, ex);
							return;
						}
					}
				}
			}, "query-thread-"+i);
		}
		for(int i=QUERY_THREADS; i<threads.length; i++){
			final int copyOfI = i;
			threads[copyOfI]=new Thread(new Runnable() {
				@Override
				public void run() {
					final int ID = 9000 + copyOfI;
					for(int j=0; j<ITERATIONS; j++){
						try {
							LoginDataType addRequest=getAddRequest(ID);
							xuudb.add(addRequest);
							
							LoginDataType queryResponse=doQuery();
							assertEquals("user" + ID, queryResponse.getRole());
							assertEquals("login" + ID, queryResponse.getXlogin());
							assertEquals("projects" + ID, queryResponse.getProjects());

							LoginDataType removeRequest=getRemoveByXloginRequest("xlogin"+ID);
							xuudb.remove(removeRequest);
						} catch (AssertionError ex) {
							errors.put(copyOfI, ex);
							return;
						} catch (Exception ex) {
							errors.put(copyOfI, ex);
							return;
						}
					}
				}
			}, "crud-thread-"+i);
		}

		long time = System.currentTimeMillis();
		for(Thread th:threads) th.start();
		for(Thread th:threads) th.join();
		time = System.currentTimeMillis() - time;
		
		for(Map.Entry<Integer,Throwable> error: errors.entrySet()){
			System.out.println("Thread #"+error.getKey()+" caused an exception: "+error.getValue().getMessage());
		}
		LoginDataType removeRequest=LoginDataType.Factory.newInstance();
		xuudb.remove(removeRequest);
		assertEquals(0, errors.size());

		System.out.println("Average speed: " +
				(ITERATIONS*1000.0*(QUERY_THREADS+CRUD_THREADS*3)/time) +
				" operations/s");
	}

	private LoginDataType[] getEntries()throws Exception{
		LoginDataType query=LoginDataType.Factory.newInstance();
		query.setGcID("foo");
		return xuudb.listDB(query);
	}
	
	protected LoginDataType getAddRequest(String login)throws Exception{
		String cert=X509Utils.getStringFromPEMFile(new File("src/test/resources/demo-user.pem"));
		LoginDataType add=LoginDataType.Factory.newInstance();
		add.setGcID("foo");
		add.setRole("user");
		add.setXlogin(login);
		add.setToken(cert);
		return add;
	}

	private LoginDataType doQuery()throws Exception{
		String cert=X509Utils.getStringFromPEMFile(new File("src/test/resources/demo-user.pem"));
		return xuudb.checkCert("foo", cert);
	}

	private LoginDataType getRemoveByXloginRequest(String login)throws Exception{
		LoginDataType add=LoginDataType.Factory.newInstance();
		add.setXlogin(login);
		return add;
	}
	

	private LoginDataType getRemoveRequest(String login)throws Exception{
		String cert=X509Utils.getStringFromPEMFile(new File("src/test/resources/demo-user.pem"));
		LoginDataType add=LoginDataType.Factory.newInstance();
		add.setGcID("foo");
		add.setXlogin(login);
		add.setToken(cert);
		return add;
	}
	
	private LoginDataType getRemoveByDNRequest()throws Exception{
		LoginDataType add=LoginDataType.Factory.newInstance();
		String cert=X509Utils.getStringFromPEMFile(new File("src/test/resources/demo-user.pem"));
		String dn=SecurityToken.getDN(cert, false);
		add.setToken(dn);
		return add;
	}


	protected LoginDataType getAddRequest(int id)throws Exception{
		LoginDataType add=LoginDataType.Factory.newInstance();
		add.setGcID("foo"+id);
		add.setRole("user"+id);
		add.setXlogin("login"+id);
		add.setProjects("projects"+id);
		add.setToken(X509Utils.getStringFromPEMFile(new File("src/test/resources/demo-user.pem")));
		return add;
	}


	private LoginDataType doQuery(String gcid)throws Exception{
		return xuudb.checkCert(gcid, X509Utils.getStringFromPEMFile(new File("src/test/resources/demo-user.pem")));
	}

}
